"""
Copyright © 2021, 2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

# Supervise:
#   Task: loop.run_in_executor(None, server.serve_forever)


import multiprocessing
import multiprocessing.connection as mpc

import pyDE1.config
from pyDE1.exceptions import *


# Right now, this is all "sync" processing. As it is a benefit to only have
# one request pending at a time, this shouldn't be a big problem.
# Going to async for the "second half" of waiting for the response
# might be a way to provide a timeout and prevent permanent blocking.


def run_api_inbound(master_config: pyDE1.config.Config,
                    log_queue: multiprocessing.Queue,
                    api_pipe: mpc.Connection):

    # Not clear why this one needs to be different
    import pyDE1.config
    pyDE1.config.config = master_config
    from pyDE1.config import config

    import asyncio
    import http.server
    import json
    import os
    import re
    import time

    from email.utils import formatdate  # RFC2822 dates
    from http import HTTPStatus
    from traceback import TracebackException
    from typing import Optional, Union, NamedTuple, Dict, Pattern

    import pyDE1
    import pyDE1.pyde1_logging as pyde1_logging
    import pyDE1.shutdown_manager as sm

    from pyDE1.dispatcher.mapping import MAPPING, mapping_requires
    from pyDE1.dispatcher.resource import Resource
    from pyDE1.dispatcher.payloads import APIRequest, APIResponse, HTTPMethod
    from pyDE1.dispatcher.validate import validate_patch_return_targets
    # These two needed as they have specific fields that need to be unpickled
    # from pyDE1.exceptions import *  # Only allowed at module level
    from pyDE1.exceptions import (
        DE1APIUnsupportedStateTransitionError, DE1APIUnsupportedFeatureError
    )
    from pyDE1.supervise import SupervisedTask, SupervisedExecutor
    from pyDE1.utils import timestamp_to_str_with_ms

    logger = pyDE1.getLogger('Inbound')

    pyde1_logging.setup_queue_logging(config.logging, log_queue)
    pyde1_logging.config_logger_levels(config.logging)

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    def on_shutdown_underway_cleanup():
        logger.info("Watching for shutdown event")
        sm.shutdown_underway.wait()
        logger.info("Shutting down HTTP server")
        try:
            server.shutdown()
        except (NameError, AttributeError):
            pass
        logger.info("Setting cleanup_complete")
        sm.cleanup_complete.set()

    on_shutdown_wait_task = loop.run_in_executor(
        None, on_shutdown_underway_cleanup)

    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    loop.set_exception_handler(sm.exception_handler)

    async def heartbeat():
        hlog = pyDE1.getLogger('Heartbeat.InboundAPI')
        while True:
            await asyncio.sleep(10)
            hlog.debug("===== BOOP =====")

    SupervisedTask(heartbeat)

    MIME_TYPE_MAP = {
        '.txt': 'text/plain',
        '.log': 'text/plain',
        '.json': 'application/json',
        '.zip': 'application/zip',
        '.gz': 'application/gz',
        '.bz2': 'application/x-bzip2',
        '.xz': 'application/x-xz',
    }

    MIME_TYPE_DEFAULT = 'application/octet-stream'

    class FileDetails(NamedTuple):
        id: str
        name: str  # Redundant
        size: int
        atime: float
        mtime: float
        ctime: float

    def file_detail_list(dirname: str):
        if not os.path.isdir(dirname):
            raise DE1TypeError(
                f"Apparent misconfiguration as '{dirname}' "
                "is not a directory")
        retval = []
        with os.scandir(dirname) as dir_entries:
            for dir_entry in dir_entries:
                dir_entry: os.DirEntry
                if not dir_entry.is_file():
                    continue
                dir_entry_stat = dir_entry.stat()
                details = FileDetails(
                    id=dir_entry.name,
                    name=dir_entry.name,
                    size=dir_entry_stat.st_size,
                    atime=dir_entry_stat.st_atime,
                    mtime=dir_entry_stat.st_mtime,
                    ctime=dir_entry_stat.st_ctime
                )
                retval.append(details._asdict())
        return retval

    # TODO: This should somehow be "automated" and driven off Resource

    resources_with_params: Dict[Resource, Pattern] = {
        Resource.LOG: re.compile('^log/(?P<id>[a-zA-Z0-9._-]+)$')
    }

    class RequestHandler (http.server.BaseHTTPRequestHandler):

        def __init__(self, *args, **kwargs):
            # Calling super().__init__() starts processing, set up logger first
            self._start = time.time()
            self._logger = pyDE1.getLogger('Inbound.HTTP')
            super(RequestHandler, self).__init__(*args, **kwargs)

        def _format_log_message(self, fmt, *args) -> str:
            # No need for CLF here
            dt = (time.time() - self._start) * 1000
            return "%.0f %s %s" % (dt,
                                   fmt % args,
                                   self.address_string())

        def log_message(self, fmt, *args):
            self._logger.info(self._format_log_message(fmt, *args))

        def log_error(self, fmt, *args):
            self._logger.error(self._format_log_message(fmt, *args))

        def log_request(self, code='-', size='-'):
            if isinstance(code, HTTPStatus):
                code = f'{code.value} "{code.phrase}"'
            self.log_message('%s %s %s',
                             str(code), str(size), self.requestline)

        def send_error_response(self,
                                code: Union[HTTPStatus, int], resp_str: str,
                                timestamp: Optional[float] = None):
            if timestamp is None:
                timestamp = time.time()
            resp_bytes = resp_str.encode('utf-8')
            self.send_response(code)
            self.send_header("Content-type", "text/plain")
            self.send_header("Content-length", str(len(resp_bytes)))
            self.send_header("Last-Modified", formatdate(timestamp,
                                                         localtime=True))
            self.end_headers()
            self.wfile.write(resp_bytes)

        def get_resource(self) -> (Optional[Resource], Dict):
            self._logger.info(f"Request: {self.requestline}")
            resource: Optional[Resource] = None
            parameter_dict = {}
            code = None
            resp_str = ''
            root_relative = self.path.removeprefix(config.http.SERVER_ROOT)
            try:
                # resource = Resource(self.path.removeprefix(SERVER_ROOT))
                resource = Resource(root_relative)
            except ValueError:
                for res, pattern in resources_with_params.items():
                    res: Resource
                    pattern: Pattern
                    match = pattern.match(root_relative)
                    if match is not None:
                        resource = res
                        parameter_dict = match.groupdict()
                    break

            if resource is None:
                code = HTTPStatus.NOT_FOUND
                resp_str = f"Unrecognized resource {self.requestline}"

            elif ((self.command == "GET" and not resource.can_get)
                  or (self.command == "PATCH" and not resource.can_patch)
                  or (self.command == "PUT" and not resource.can_put)
                  or (self.command == "POST" and not resource.can_post)
                  or (self.command == "DELETE" and not resource.can_delete)
                  or self.command not in (
                          'GET', 'PATCH', 'PUT', 'POST', 'DELETE')):
                code = HTTPStatus.METHOD_NOT_ALLOWED
                resp_str = f"{self.command} not permitted for {resource}"

            if code is not None:
                self.send_error_response(code, resp_str)

            return resource, parameter_dict

        # NB: This does not support Transfer-encoding: chunked
        def get_content(self) -> Optional[Union[bytes, bytearray, str]]:

            content = None
            content_length = self.headers.get('content-length')

            if content_length is None:
                self.send_error_response(
                    HTTPStatus.LENGTH_REQUIRED,
                    "Missing Content-Length header")
                return None

            else:
                content_length = int(content_length)


            this_content_limit = config.http.PATCH_SIZE_LIMIT
            try:
                resource = Resource(self.path.removeprefix(
                                                config.http.SERVER_ROOT))
                if resource == Resource.DE1_FIRMWARE:
                    this_content_limit = 1 * 1024 * 1024  # FW1258 < 500 kB
            except ValueError:
                pass

            if content_length > this_content_limit:
                self.send_error_response(
                    HTTPStatus.REQUEST_ENTITY_TOO_LARGE,
                    "Patch is too large")

            else:
                if content_length > 0:
                    content = self.rfile.read(content_length)
                else:
                    content = ''

            return content

        def queue_and_respond(self, req: APIRequest):

            api_pipe.send(req)

            # In the sync world, nothing in parallel in this process
            # so might as well just block, rather than craziness

            readable = api_pipe.poll(timeout=config.http.RESPONSE_TIMEOUT)
            if readable:
                resp = api_pipe.recv()
            else:
                e = TimeoutError(
                    "Timeout waiting for response from controller, "
                    f"over {config.http.RESPONSE_TIMEOUT} sec")
                resp = APIResponse(
                    timestamp=time.time(),
                    original_timestamp=req.timestamp,
                    payload=None,
                    exception=e,
                    tbe=TracebackException.from_exception(e)
                )

            self.process_response(resp)

        # Split as some requests are handled in this process directly
        def process_response(self, resp: APIResponse,
                             mime_type: 'str' = "application/json"):

            if resp.exception is None:
                content = resp.payload
                if mime_type.endswith('/json') and not isinstance(content, str):
                    content = json.dumps(content,
                                          sort_keys=True, indent=4) + "\n"
                if isinstance(content, str):
                    content = content.encode('utf-8')

                self.send_response(HTTPStatus.OK)
                self.send_header("Content-type", mime_type)
                self.send_header("Content-length", str(len(content)))
                self.send_header("Last-Modified", formatdate(resp.timestamp,
                                                             localtime=True))
                self.end_headers()
                self.wfile.write(content)

            else:

                body = ''.join(resp.tbe.format())

                if isinstance(resp.exception,
                              (DE1DBNoMatchingRecord,)):
                    http_status = HTTPStatus.NOT_FOUND

                elif isinstance(resp.exception,
                              (DE1APIUnsupportedStateTransitionError,
                               DE1NotConnectedError,
                               DE1IsConnectedError,
                               DE1NoAddressError,
                               DE1OperationInProgressError,)):
                    http_status = HTTPStatus.CONFLICT

                elif isinstance(resp.exception,
                                DE1APIUnsupportedFeatureError):
                    http_status = HTTPStatus.IM_A_TEAPOT

                elif isinstance(resp.exception, DE1APIError):
                    http_status = HTTPStatus.BAD_REQUEST

                elif isinstance(resp.exception,
                                (TimeoutError,
                                 asyncio.exceptions.TimeoutError)):
                    http_status = HTTPStatus.REQUEST_TIMEOUT

                else:
                    http_status = HTTPStatus.INTERNAL_SERVER_ERROR

                self.send_error_response(code=http_status,
                                         resp_str=body,
                                         timestamp=resp.timestamp)

            rtt = (time.time() - resp.original_timestamp) * 1000
            self._logger.debug(
                f"RTT: {rtt:0.1f} ms {self.requestline}"
            )
            return

        def do_GET(self):

            timestamp = time.time()
            (resource, parameter_dict) = self.get_resource()
            if resource is None:
                return

            if resource == Resource.LOGS:
                payload = None
                exc = None
                tbe = None
                try:
                    payload = file_detail_list(config.logging.LOG_DIRECTORY)
                except Exception as e:
                    exc = e
                    tbe = TracebackException.from_exception(e)

                resp = APIResponse(
                    original_timestamp=timestamp,
                    timestamp=time.time(),
                    payload=payload,
                    exception=exc,
                    tbe=tbe)

                self.process_response(resp)

            elif resource == Resource.LOG:

                payload = None
                exc = None
                tbe = None

                # TODO: Another ugly combination of id with filename
                filename = os.path.join(config.logging.LOG_DIRECTORY,
                                        parameter_dict['id'])

                try:
                    with open(filename, 'rb') as log_file:
                        payload = log_file.read()
                except Exception as e:
                    exc = e
                    tbe = TracebackException.from_exception(e)

                resp = APIResponse(
                    original_timestamp=timestamp,
                    timestamp=time.time(),
                    payload=payload,
                    exception=exc,
                    tbe=tbe)

                mime_type = MIME_TYPE_DEFAULT
                for suffix, mime_for_suffix in MIME_TYPE_MAP.items():
                    if filename.endswith(suffix):
                        mime_type = mime_for_suffix
                        break

                self.process_response(resp, mime_type)

            else:

                # Not actionable here as connectivity is unknown
                requires = mapping_requires(MAPPING[resource])

                req = APIRequest(timestamp=timestamp,
                                 method=HTTPMethod.GET,
                                 resource=resource,
                                 connectivity_required=requires,
                                 payload=None)

                self.queue_and_respond(req)

            return

        def do_PATCH(self):

            timestamp = time.time()
            (resource, parameter_dict) = self.get_resource()
            if resource is None:
                return

            content = self.get_content()
            if content is None:
                self.send_error_response(
                    HTTPStatus.BAD_REQUEST,
                    "No content provided for PATCH request"
                )
                return

            try:
                patch = json.loads(content)
                targets = validate_patch_return_targets(resource=resource,
                                                        patch=patch)

            except (json.JSONDecodeError, DE1APIError) as exception:
                self.send_error_response(
                    HTTPStatus.BAD_REQUEST,
                    repr(exception))
                return

            req = APIRequest(timestamp=timestamp,
                             method=HTTPMethod.PATCH,
                             resource=resource,
                             connectivity_required=targets,
                             payload=patch)

            self.queue_and_respond(req)
            return

        def do_PUT(self):

            timestamp = time.time()
            (resource, parameter_dict) = self.get_resource()
            if resource is None:
                return

            if resource not in (Resource.DE1_PROFILE,
                                Resource.DE1_PROFILE_ID,
                                Resource.DE1_PROFILE_STORE,
                                Resource.DE1_FIRMWARE,
                                Resource.DE1_FIRMWARE_CANCEL):
                self.send_error_response(
                    HTTPStatus.NOT_IMPLEMENTED,
                    f"PUT not yet supported for {resource}"
                )
                return

            content = self.get_content()
            if content is None:
                self.send_error_response(
                    HTTPStatus.BAD_REQUEST,
                    "No content provided for PUT request"
                )
                return

            try:
                if resource in (Resource.DE1_PROFILE,
                                Resource.DE1_PROFILE_STORE,
                                Resource.DE1_FIRMWARE,
                                Resource.DE1_FIRMWARE_CANCEL,
                                ):
                    patch = content
                else:
                    patch = json.loads(content)
                targets = validate_patch_return_targets(resource=resource,
                                                        patch=patch)

            except (json.JSONDecodeError, DE1APIError) as exception:
                self.send_error_response(
                    HTTPStatus.BAD_REQUEST,
                    repr(exception))
                return

            req = APIRequest(timestamp=timestamp,
                             method=HTTPMethod.PUT,
                             resource=resource,
                             connectivity_required=targets,
                             payload=patch)

            self.queue_and_respond(req)
            return

    # try:
    server = http.server.HTTPServer((config.http.SERVER_HOST,
                                     config.http.SERVER_PORT),
                                    RequestHandler)

    # Not clear why execution continues even with
    #   OSError: [Errno 98] Address already in use

    # As this may be a restart, ensure that there are not pending responses
    while api_pipe.poll():
        got = api_pipe.recv()
        try:
            from_str = f" from{timestamp_to_str_with_ms(got.timestamp)}"
        except AttributeError:
            from_str = ''
        logger.warning(f"Flushing stale response{from_str}: {got}")

    SupervisedExecutor(None, server.serve_forever)

    loop.run_forever()



"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only


FlowSequencer

Responsible for coordinating the actions around any of the flow modes
"""
import asyncio
import enum
import multiprocessing
import time
from asyncio import Task
from typing import Optional, Callable, Coroutine, List, Union, NamedTuple

import pyDE1
from pyDE1.config import config
from pyDE1.database.create_history_record import create_history_record
from pyDE1.database.recorder_control import RecorderControl
from pyDE1.de1 import DE1
from pyDE1.de1.ble import CUUID
from pyDE1.de1.c_api import (
    API_MachineStates, API_Substates, MMR0x80LowAddr, ShotSample
)
from pyDE1.de1.events import (
    ShotSampleUpdate, StateUpdate, ShotSampleWithVolumesUpdate
)
from pyDE1.event_manager import (
    SequencerGate, SequencerGateName, SequencerGateNotification,
    send_to_outbound_pipes
)
from pyDE1.exceptions import (
    DE1APIValueError, DE1APIAttributeError, DE1APITypeError,
    DE1APINotManagedHereException
)
from pyDE1.flow_sequencer import (
    FlowSequencer, LAST_DROPS_MINIMUM_TIME_DEFAULT,
    FIRST_DROPS_THRESHOLD_DEFAULT, StopAtNotificationAction, StopAtType,
    StopAtNotification, AutoTareNotificationAction, AutoTareNotification,
    ModeControl, BaseModeControl, validate_stop_at,
    StopAtTimeControl, StopAtVolumeControl, StopAtWeightControl,
    MoveOnWeightControl,
    I_EspressoControl, I_HotWaterControl,
    I_HotWaterRinseControl, I_SteamControl,
)
from pyDE1.scale.events import WeightAndFlowUpdate
from pyDE1.scale.processor import ScaleProcessor
from pyDE1.singleton import Singleton
from pyDE1.utils import cancel_tasks_by_name

logger = pyDE1.getLogger('FlowSequencer')


class FlowSequencerImpl (Singleton, FlowSequencer):

    # NB: This is intentionally done in _singleton_init() and not __init__()
    #     See Singleton and Guido's notes there
    #
    # def __init__(self):
    #     pass

    database_queue: Optional[multiprocessing.Queue] = None

    def _singleton_init(self):

        # Singletons now
        self._de1 = DE1()
        self._scale_processor = ScaleProcessor()

        # self._event_de1_present = asyncio.Event()
        # self._event_scale_processor_present = asyncio.Event()

        self._active_state: Optional[API_MachineStates] = None

        # Detect frame changes for move-on by weight
        self._last_profile_frame = None
        self._last_profile_frame_weight = None
        self._last_frame_advanced_from = None

        self._sequence_start_time: float = 0
        # For later cancellation. Should this be static or dynamic?
        self._sequence_task_name: str = "SequencerSubtask"

        self._gate_sequence_start = SequencerGate(self,
            SequencerGateName.GATE_SEQUENCE_START)
        # These are not necessarily in order
        self._gate_flow_begin = SequencerGate(self,
            SequencerGateName.GATE_FLOW_BEGIN)
        self._gate_expect_drops = SequencerGate(self,
            SequencerGateName.GATE_EXPECT_DROPS)
        # self._gate_first_drops = asyncio.Event()
        self._gate_exit_preinfuse = SequencerGate(self,
            SequencerGateName.GATE_EXIT_PREINFUSE)
        self._gate_flow_end = SequencerGate(self,
            SequencerGateName.GATE_FLOW_END)
        self._gate_flow_state_exit = SequencerGate(self,
            SequencerGateName.GATE_FLOW_STATE_EXIT)
        self._gate_last_drops = SequencerGate(self,
            SequencerGateName.GATE_LAST_DROPS)
        # Though always "done" here (and should always be triggered)
        self._gate_sequence_complete = SequencerGate(self,
            SequencerGateName.GATE_SEQUENCE_COMPLETE)

        self._all_gates = [
            self._gate_sequence_start,
            self._gate_flow_begin,
            self._gate_expect_drops,
            self._gate_exit_preinfuse,
            self._gate_flow_end,
            self._gate_flow_state_exit,
            self._gate_last_drops,
            self._gate_sequence_complete,
        ]

        self.espresso_control = EspressoControl()
        self.steam_control = SteamControl()
        self.hot_water_control = HotWaterControl()
        self.hot_water_rinse_control = HotWaterRinseControl()

        # TODO: Refactor these out in favor of None in the ByModeControl
        self.autotare_states: List[API_MachineStates] = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,
        ]

        self.stop_at_weight_states: List[API_MachineStates] = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,
        ]

        self.move_on_weight_states: List[API_MachineStates] = [
            API_MachineStates.Espresso,
        ]

        self.stop_at_volume_states: List[API_MachineStates] = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,
        ]

        self.last_drops_states = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,     # TODO: Needed? Adjust?
            API_MachineStates.Steam,        # TODO: Needed? Adjust?
        ]

        self.stop_at_time_states = [
            API_MachineStates.HotWaterRinse,
            # API_MachineStates.Steam,  # Handled through ShotSettings
        ]

        # Recorder gets managed by the recorder itself

        # Internal flag, use None or value for
        # de1.stop_at_weight and/or de1.stop_at_volume
        self._stop_at_weight_active = False
        self._move_on_weight_active = False
        self._stop_at_volume_active = False
        self._stop_at_time_active = False

        self._sequence_watchdog_task: Optional[Task] = None

        asyncio.create_task(self.set_up_subscribers())

    @property
    def stop_at_weight_adjust(self):
        """
        In seconds, larger increases weight in cup

        Supply as a property to simplify the stop-at-weight closure
        and allow the value to be changed after the closure is created
        """
        return config.de1.STOP_AT_WEIGHT_ADJUST

    @property
    def de1(self):
        return self._de1

    async def set_up_subscribers(self):
        # DE1 only knows that its _flow_sequencer is an I_TargetSetter
        # because of circular imports, so it can't reference FlowSequencer
        self._de1._flow_sequencer = self

        await asyncio.gather(
            self.de1.event_state_update.subscribe(
                self._create_state_update_callback()),

            self.de1.event_shot_sample.subscribe(
                self._create_shot_sample_update_callback()),

            self.de1.event_shot_sample_with_volumes_update.subscribe(
                self._create_stop_at_volume_subscriber()),

            self.scale_processor.event_weight_and_flow_update.subscribe(
                self._create_act_on_weight_subscriber()),
        )
        logger.info("FlowSequencer subscriptions done")
        return self

    @property
    def scale_processor(self) -> ScaleProcessor:
        return self._scale_processor

    @property
    def active_state(self) -> API_MachineStates:
        return self._active_state

    @property
    def current_frame(self) -> Optional[int]:
        # Tighten up a bit from DE1
        if (self.active_state is API_MachineStates.Espresso
                and self.de1.current_substate.flow_phase == 'during'):
            retval = self.de1.current_frame
        else:
            retval = None
        return retval

    def active_control_for_state(self, state: API_MachineStates) -> ModeControl:
        retval = None
        if state is API_MachineStates.Espresso:
            retval = self.espresso_control
        elif state is API_MachineStates.Steam:
            retval = self.steam_control
        elif state is API_MachineStates.HotWater:
            retval = self.hot_water_control
        elif state is API_MachineStates.HotWaterRinse:
            retval = self.hot_water_rinse_control
        return retval

    @property
    def active_control(self) -> ModeControl:
        return self.active_control_for_state(self.active_state)

    @property
    def active_control_stop_at_time(self):
        try:
            retval = self.active_control.stop_at_time
        except AttributeError:
            retval = None
        return retval

    @property
    def active_control_stop_at_volume(self):
        try:
            retval = self.active_control.stop_at_volume
        except AttributeError:
            retval = None
        return retval

    @property
    def active_control_stop_at_weight(self):
        try:
            retval = self.active_control.stop_at_weight
        except AttributeError:
            retval = None
        return retval

    @property
    def sequence_start_time(self):
        return self._sequence_start_time

    @property
    def sequence_is_running(self):
        return (not self._gate_sequence_complete.is_set()
                and self._sequence_start_time != 0)

    @property
    def sequence_is_running_and_flow_not_started(self):
        return (self.sequence_is_running
                and not self._gate_flow_begin.is_set())

    def _close_all_gates(self):
        for gate in self._all_gates:
            gate.clear()

    # Weight and Flow Update handles auto-tare
    # SAW and SAV have their own handlers, supervisor enables and disables
    # Subscribe to the ScaleProcessor, let it handle scale-instance changes

    def _create_weight_and_flow_update_callback(self) -> Callable:
        flow_sequencer = self

        async def flow_sequencer_wafu_cb(wafu: WeightAndFlowUpdate):
            nonlocal flow_sequencer
            pass  # SAW is _create_act_on_weight_subscriber

        return flow_sequencer_wafu_cb

    def _create_state_update_callback(self) -> Callable:
        flow_sequencer = self

        async def flow_sequencer_su_cb(su: StateUpdate):
            """
            There is the possibility that the transitions never go through flow
            such as hitting "start" and deciding to stop before flow begins.
            If this is not accounted for, the sequence might "never end".
            """
            nonlocal flow_sequencer

            if su.state.is_flow_state:

                if not su.previous_state.is_flow_state:
                    flow_sequencer._start_sequence(su.state)
                    logger.info("Start sequence")

                if su.substate.flow_phase == 'during' \
                        and su.previous_substate.flow_phase != 'during':
                    flow_sequencer._gate_flow_begin.set()
                    logger.info("Gate: Flow begin")

                if su.previous_substate == API_Substates.PreInfuse \
                        and su.substate != API_Substates.PreInfuse:
                    flow_sequencer._gate_exit_preinfuse.set()
                    logger.info("Gate: Exit preinfuse")

                if su.previous_substate.flow_phase == 'during' \
                        and su.substate.flow_phase != 'during':
                    flow_sequencer._gate_flow_end.set()
                    logger.info("Gate: Flow end")

            elif su.previous_state.is_flow_state:  # current is not a flow state
                if su.previous_substate == API_Substates.PreInfuse:
                    flow_sequencer._gate_exit_preinfuse.set()
                    logger.info("Gate: Exit preinfuse")
                if su.previous_substate.flow_phase == 'during':
                    flow_sequencer._gate_flow_end.set()
                    logger.info("Gate: Flow end")
                flow_sequencer._gate_flow_state_exit.set()
                logger.info("Gate: Flow state exit")
                if flow_sequencer.sequence_is_running_and_flow_not_started:
                    # Catch leaving before flow starts
                    logger.info(
                        "Non-flow state detected during sequence: "
                        f"{su.state.name},{su.substate.name} "
                        'calling _end_sequence().')
                    await flow_sequencer._end_sequence()

        return flow_sequencer_su_cb

    def _create_shot_sample_update_callback(self) -> Callable:
        flow_sequencer = self

        async def flow_sequencer_ssu_cb(ssu: ShotSampleUpdate):
            nonlocal flow_sequencer

            if ssu.frame_number != self._last_profile_frame:
                logger.info(
                    "Frame change "
                    f"{self._last_profile_frame} => {ssu.frame_number} at "
                    f"{self.scale_processor.current_weight} g")
                self._last_profile_frame = ssu.frame_number
                self._last_profile_frame_weight = self.scale_processor.current_weight

            try:
                threshold = flow_sequencer.active_control.first_drops_threshold
            except AttributeError:
                threshold = None
            if self.de1.current_state.is_flow_state \
                    and threshold is not None \
                    and not flow_sequencer._gate_expect_drops.is_set() \
                    and self.de1.current_substate.flow_phase == 'during' \
                    and ssu.group_pressure \
                        > flow_sequencer.active_control.first_drops_threshold:
                flow_sequencer._gate_expect_drops.set()
                logger.info("Gate: Expect drops")

        return flow_sequencer_ssu_cb

    # Should this be async def now?
    def _start_sequence(self, state: API_MachineStates):
        """
        Kick off parallel tasks to manage functions during the shot
        Tasks should wait for self._gate_sequence_start
        """
        self._sequence_start_time = time.time()
        # TODO: Is there a more robust way to transition from an "aborted"
        #       shot to another one?
        if self.sequence_is_running:
            self._abort_sequence()

        self._active_state = state

        sequence_id = SequencerGateNotification.new_sequence()
        logger.info(f"Starting {self.active_state.name}, "
                    f"sequence_id {sequence_id}")

        # Can't really start the recorder here as it needs
        # to be able to create the sequence (history) record
        # which can take up to 250 ms

        self._close_all_gates()

        # create a bunch of tasks
        # Keep the start gate closed, as some may expect others to exist
        # or need to
        # For now, don't name uniquely and see what the logger returns

        asyncio.create_task(self._sequence_end_sequence(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_volume_tracking(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_stop_at_volume(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_scale_tare(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_stop_at_weight(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_move_on_weight(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_stop_at_time(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_recorder(),
                            name=self._sequence_task_name)

        self._sequence_watchdog_task = asyncio.create_task(
            self._sequence_watchdog(),
            name=f"SequenceWatchdog_{config.de1.SEQUENCE_WATCHDOG_TIMEOUT}")

        self._last_profile_frame_weight = None
        self._last_frame_advanced_from = None

        self._gate_sequence_start.set()
        logger.info("Gate: Sequence start")

    async def _sequence_end_sequence(self):
        """
        Maybe a funky way to do it, but this captures "done-ness"
        in one place.
        """
        await self._gate_flow_end.wait()
        try:
            ldmt = self.active_control.last_drops_minimum_time
        except AttributeError:
            ldmt = None
        if ldmt:
            try:
                await asyncio.wait_for(self._gate_last_drops.wait(),
                                  timeout=ldmt)
            except asyncio.exceptions.TimeoutError:
                pass
        self._gate_last_drops.set()
        await self._gate_flow_state_exit.wait()
        await self._end_sequence()

    async def _end_sequence(self):
        """
        Not all subtasks will have gone through all their gates
        Set the completion Event and give a second for tasks to complete
        Clean up any stragglers
        """
        if self._sequence_watchdog_task is not None:
            self._sequence_watchdog_task.cancel("by _end_sequence()")
        self._gate_sequence_complete.set()
        logger.info("Gate: Sequence complete")
        await asyncio.sleep(0.100)
        self._cleanup_stragglers()

    def _abort_sequence(self):
        """
        Not all subtasks will have gone through all their gates
        or sequence_complete may not have been set.

        This is a "no-wait" version

        Set the completion Event
        Clean up any stragglers
        """
        if self._sequence_watchdog_task is not None:
            self._sequence_watchdog_task.cancel("by _abort_sequence()")
        self._gate_sequence_complete.set()
        logger.info("Gate: Sequence complete: _abort_sequence()")
        self._cleanup_stragglers()

    def _cleanup_stragglers(self):
        """
        This will send CancelledError to the coroutine

        That should be caught and handled for cleanup
        or completion of tasks, such as writing files

        cancel_tasks_by_name() won't cancel its own task
        """
        logger.debug("Canceling stragglers")
        cancel_tasks_by_name(self._sequence_task_name, starts_with=False)


###
### Sequences
###

    # TODO: These should be sensitive to the kind of flow
    #       Just get them in for espresso for now

    async def _sequence_volume_tracking(self):
        de1 = self.de1
        try:
            await self._gate_sequence_start.wait()
            de1._reset_volume_dispensed()
            logger.debug("Volume: reset")

            await self._gate_flow_begin.wait()
            de1._tracking_volume_dispensed = True
            logger.debug("Volume: tracking start")

            await self._gate_flow_end.wait()
            de1._tracking_volume_dispensed = False
            logger.debug("Volume: tracking stop")

        except asyncio.CancelledError:
            de1._tracking_volume_dispensed = False
            logger.info("Volume: tracking stop - on cancel")
            raise

    async def _sequence_scale_tare(self):
        scale = self._scale_processor.scale
        if scale is None:
            logger.info("No scale present, will not hold at tare")
            return
        if self.active_state not in self.autotare_states:
            scale.hold_at_tare = False
            logger.debug(f"Scale: release - {self.active_state.name}")
            await self.auto_tare_notify(AutoTareNotificationAction.DISABLED)
            return
        try:
            await self._gate_sequence_start.wait()
            scale.hold_at_tare = True
            logger.debug("Scale: hold at tare")
            await self.auto_tare_notify(AutoTareNotificationAction.ENABLED)

            await self._gate_expect_drops.wait()
            scale.hold_at_tare = False
            logger.debug("Scale: release")
            await self.auto_tare_notify(AutoTareNotificationAction.DISABLED)

        except asyncio.CancelledError:
            scale = self._scale_processor.scale
            scale.hold_at_tare = False
            logger.info("Scale: release - on cancel")
            await self.auto_tare_notify(AutoTareNotificationAction.DISABLED)
            raise

    async def _sequence_stop_at_volume(self):
        if self.active_state not in self.stop_at_volume_states:
            self._stop_at_volume_active = False
            logger.debug(f"StopAtVolume: disable - {self.active_state.name}")
            await self.stop_at_notify(
                    stop_at=StopAtType.VOLUME,
                    action=StopAtNotificationAction.DISABLED,
                    target_value=self.active_control_stop_at_volume,
                    current_value=None,
                    active_state=self.active_state,
                    current_frame=self.current_frame)
            return
        try:
            await self._gate_sequence_start.wait()
            self._stop_at_volume_active = False
            logger.debug("StopAtVolume: disable")
            await self.stop_at_notify(
                    stop_at=StopAtType.VOLUME,
                    action=StopAtNotificationAction.DISABLED,
                    target_value=None,
                    current_value=self.de1.volume_dispensed_pour,
                    active_state=self.active_state,
                    current_frame=self.current_frame)

            await self._gate_exit_preinfuse.wait()
            self._stop_at_volume_active = True
            logger.debug("StopAtVolume: enable")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.ENABLED,
                target_value=self.active_control_stop_at_volume,
                current_value=self.de1.volume_dispensed_pour,
                active_state=self.active_state,
                current_frame=self.current_frame)

            await self._gate_flow_end.wait()
            self._stop_at_volume_active = False
            logger.debug("StopAtVolume: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_volume,
                current_value=self.de1.volume_dispensed_pour,
                active_state=self.active_state,
                current_frame=self.current_frame)

        except asyncio.CancelledError:
            self._stop_at_volume_active = False
            logger.info("StopAtVolume: disable - on cancel")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_volume,
                current_value=self.de1.volume_dispensed_pour,
                active_state=self.active_state,
                current_frame=self.current_frame)
            raise

    async def _sequence_stop_at_weight(self):
        if self.active_state not in self.stop_at_weight_states:
            self._stop_at_weight_active = False
            logger.info(f"StopAtWeight: disable - {self.active_state.name}")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_weight,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)
            return
        try:
            await self._gate_sequence_start.wait()
            self._stop_at_weight_active = False
            logger.debug("StopAtWeight: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_weight,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)

            await self._gate_expect_drops.wait()
            self._stop_at_weight_active = True
            logger.debug("StopAtWeight: enable")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.ENABLED,
                target_value=self.active_control_stop_at_weight,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)

            await self._gate_flow_end.wait()
            self._stop_at_weight_active = False
            logger.debug("StopAtWeight: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_weight,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)

        except asyncio.CancelledError:
            self._stop_at_weight_active = False
            logger.info("StopAtWeight: disable - on cancel")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_weight,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)
            raise

    async def _sequence_move_on_weight(self):
        """
        At this level, there really isn't a meaningful "target value"
        as it is a list of values, potentially for each frame
        """
        if self.active_state not in self.move_on_weight_states:
            self._move_on_weight_active = False
            logger.info(f"StopAtWeight: disable - {self.active_state.name}")
            await self.stop_at_notify(
                stop_at=StopAtType.MOW,
                action=StopAtNotificationAction.DISABLED,
                target_value=None,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)
            return
        try:
            await self._gate_sequence_start.wait()
            self._move_on_weight_active = False
            logger.debug("StopAtWeight: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.MOW,
                action=StopAtNotificationAction.DISABLED,
                target_value=None,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)

            await self._gate_expect_drops.wait()
            self._move_on_weight_active = True
            logger.debug("StopAtWeight: enable")
            await self.stop_at_notify(
                stop_at=StopAtType.MOW,
                action=StopAtNotificationAction.ENABLED,
                target_value=None,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)

            await self._gate_flow_end.wait()
            self._move_on_weight_active = False
            logger.debug("StopAtWeight: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.MOW,
                action=StopAtNotificationAction.DISABLED,
                target_value=None,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)

        except asyncio.CancelledError:
            self._move_on_weight_active = False
            logger.info("StopAtWeight: disable - on cancel")
            await self.stop_at_notify(
                stop_at=StopAtType.MOW,
                action=StopAtNotificationAction.DISABLED,
                target_value=None,
                current_value=self.scale_processor.current_weight,
                active_state=self.active_state,
                current_frame=self.current_frame)
            raise

    async def _sequence_stop_at_time(self):
        t0 = 0   # Make sure always set so task cancellation doesn't error
        if ((self.active_state == API_MachineStates.Steam)
                or (self.active_state == API_MachineStates.HotWaterRinse
                    and self.de1.feature_flag.rinse_control)):
            self._stop_at_time_active = False
            logger.info(
                "StopAtTime: DE1 controlled for "
                f"{self.active_state.name}  "
                f"({self.active_control.stop_at_time:.1f} seconds)")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DE1CONTROLLED,
                target_value=self.active_control_stop_at_time,
                current_value=None,
                active_state=self.active_state,
                current_frame=self.current_frame)
            return

        if self.active_state not in self.stop_at_time_states:
            self._stop_at_time_active = False
            logger.info(f"StopAtTime: disable - {self.active_state.name}")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_time,
                current_value=None,
                active_state=self.active_state,
                current_frame=self.current_frame)
            return

        try:
            await self._gate_sequence_start.wait()
            self._stop_at_time_active = False
            logger.debug("StopAtTime: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_time,
                current_value=None,
                active_state=self.active_state,
                current_frame=self.current_frame)

            # NB: Changing the time after starting won't alter the duration
            #     at least as presently implemented

            try:
                wait = self.active_control.stop_at_time
            except DE1APINotManagedHereException as e:
                # Steam timeout is managed by the DE1 itself
                # though shouldn't get here
                logger.error(
                    f"Unexpectedly getting through to {e}, "
                    "check stop_at_time_states")
                return

            if wait is None or wait <= 0:
                return

            await self._gate_flow_begin.wait()
            self._stop_at_time_active = True
            t0 = time.time()
            logger.debug(f"StopAtTime: enable ({wait} seconds)")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.ENABLED,
                target_value=self.active_control_stop_at_time,
                current_value=0,
                active_state=self.active_state,
                current_frame=self.current_frame)

            await asyncio.sleep(wait)
            logger.debug(f"StopAtTime: triggered, requesting stop_flow")
            await self.de1.stop_flow()
            logger.debug(f"StopAtTime: triggered ({wait} seconds)")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.TRIGGERED,
                target_value=self.active_control_stop_at_time,
                current_value=(time.time() - t0),
                active_state=self.active_state,
                current_frame=self.current_frame)

            await self._gate_flow_end.wait()
            self._stop_at_time_active = False
            logger.debug("StopAtTime: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_time,
                current_value=(time.time() - t0),
                active_state=self.active_state,
                current_frame=self.current_frame)

        except asyncio.CancelledError:
            self._stop_at_time_active = False
            logger.info("StopAtTime: disable - on cancel")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED,
                target_value=self.active_control_stop_at_time,
                current_value=((time.time() - t0) if t0 else None),
                active_state=self.active_state,
                current_frame=self.current_frame)
            raise

    async def _sequence_recorder(self):
        # Always enable recording, let the recorder decide

        de1 = self._de1

        try:
            profile_id = self.de1.latest_profile.id
        except AttributeError:
            logger.warning("Unable to determine profile for sequence record")
            profile_id = None

        try:
            await create_history_record(
                    self.active_state,
                    self.sequence_start_time,
                    profile_id)
            await self._gate_sequence_start.wait()

            FlowSequencer.database_queue.put_nowait(
                RecorderControl(
                    recording = True,
                    sequence_id=SequencerGateNotification.sequence_id))
            logger.debug("Recorder: enable")

            await self._gate_sequence_complete.wait()
            FlowSequencer.database_queue.put_nowait(
                RecorderControl(
                    recording = False,
                    sequence_id=SequencerGateNotification.sequence_id))
            logger.debug("Recorder: disable")

        except asyncio.CancelledError:
            FlowSequencer.database_queue.put_nowait(
                RecorderControl(
                    recording = False,
                    sequence_id=SequencerGateNotification.sequence_id))
            logger.info("Recorder: disable - on cancel")
            raise

    async def _sequence_watchdog(self):
        """
        Apparently there are situations where a sequence doesn't "complete"
        resulting in the recorder running until the next sequence starts.
        This can result in a bloated database.

        Changes in the state-update handler should now catch if the sequence
        is stopped before flow begins. Hopefully, there aren't other causes,
        but if there are, or if something strange happens, this timeout
        will help keep the database from recording all day.
        """
        try:
            await asyncio.sleep(wt := config.de1.SEQUENCE_WATCHDOG_TIMEOUT)
            logger.warning(
                f"Sequence watchdog timeout {wt} sec, calling _end_sequence()")
            # Don't cancel self if already running in _end_sequence()
            self._sequence_watchdog_task = None
            await self._end_sequence()
        except asyncio.CancelledError:
            pass
        finally:
            self._sequence_watchdog_task = None

    def _create_stop_at_volume_subscriber(self) -> Coroutine:
        """
        Should be subscribed to ShotSampleWithVolumesUpdate on DE1
        """
        flow_sequencer = self
        sav_logger = pyDE1.getLogger('FlowSequencer.StopAtVolume')

        async def stop_at_volume_subscriber(sswvu: ShotSampleWithVolumesUpdate):
            nonlocal flow_sequencer, sav_logger

            if (flow_sequencer._stop_at_volume_active
                    and (target := flow_sequencer.active_control.stop_at_volume)
                    is not None):
                if sswvu.volume_pour >= target:
                    await flow_sequencer.de1.stop_flow()
                    sav_logger.info(
                        "Triggered at {:.1f} mL for {:.1f} mL target".format(
                            sswvu.volume_pour, target))
                    await flow_sequencer.stop_at_notify(
                        stop_at=StopAtType.VOLUME,
                        action=StopAtNotificationAction.TRIGGERED,
                        target_value=self.active_control_stop_at_volume,
                        current_value=sswvu.volume_pour,
                        active_state=self.active_state,
                        current_frame=self.current_frame)

        return stop_at_volume_subscriber

    def _select_weight_flow (self,
                             wafu: WeightAndFlowUpdate) -> 'WeightFlowTuple':
        """
        Pick the appropriate weight and flow estimates to use

        Right now, use average unless high or negative flow suggests a bump
        Also implement "median weight always", for high-vibration settings
        """
        flow = wafu.average_flow
        flow_time = wafu.average_flow_time
        weight = wafu.current_weight
        weight_time = wafu.current_weight_time
        if config.de1.bump_resist.USE_MEDIAN_WEIGHT_ALWAYS:
            weight = wafu.median_weight
            weight_time = wafu.median_weight_time
        if config.de1.bump_resist.USE_MEDIAN_FLOW_ALWAYS:
            flow = wafu.median_flow
            flow_time = wafu.median_flow_time
        if flow > config.de1.bump_resist.FLOW_THRESHOLD:
            flow = (self.de1._cuuid_dict[
                            CUUID.ShotSample].last_value.GroupFlow
                        * config.de1.bump_resist.FLOW_MULTIPLIER)
            flow_time = self.de1._cuuid_dict[
                            CUUID.ShotSample].last_value.arrival_time
            if config.de1.bump_resist.SUB_MEDIAN_WEIGHT:
                weight = wafu.median_weight
                weight_time = wafu.median_weight_time
        elif flow < 0:
            flow = 0

        return WeightFlowTuple(
            weight=weight,
            weight_time=weight_time,
            flow=flow,
            flow_time=flow_time
        )

    def _create_act_on_weight_subscriber(self) -> Coroutine:
        """
        Should be subscribed to WeightAndFlowUpdate on ScaleProcessor
        """
        flow_sequencer = self
        saw_logger = pyDE1.getLogger('FlowSequencer.StopAtWeight')
        mow_logger = pyDE1.getLogger('FlowSequencer.MoveOnWeight')

        async def act_on_weight_subscriber(wafu: WeightAndFlowUpdate):
            nonlocal flow_sequencer, saw_logger, mow_logger

            done = False

            wft = self._select_weight_flow(wafu)

            if (flow_sequencer._stop_at_weight_active
                    and (target := flow_sequencer.active_control.stop_at_weight)
                    is not None):
                # TODO: Should the choice of flow estimate be switchable?
                dw = target - wft.weight
                if wft.flow > 0:
                    dt = dw / wft.flow
                    target_time = wft.weight_time + dt \
                                  - flow_sequencer.de1.stop_lead_time \
                                  - flow_sequencer.de1.fall_time \
                                  + flow_sequencer.stop_at_weight_adjust
                    if time.time() >= target_time:
                        await flow_sequencer.de1.stop_flow()
                        saw_logger.info(
                            "Triggered at {:.1f} g for {:.1f} g target".format(
                                wft.weight, target))
                        await flow_sequencer.stop_at_notify(
                            stop_at=StopAtType.WEIGHT,
                            action=StopAtNotificationAction.TRIGGERED,
                            target_value=target,
                            current_value=wafu.current_weight,
                            active_state=self.active_state,
                            current_frame=self.current_frame)
                        done = True

            if not done and (flow_sequencer._move_on_weight_active
                    and ((target := flow_sequencer.active_control.mow_get_frame(
                         frame := flow_sequencer.current_frame )) is not None)
                    and (frame != self._last_frame_advanced_from)):
                flow = self._adjust_flow_for_prediction(wafu.average_flow)
                start_of_frame_weight = self._last_profile_frame_weight or 0
                dw = target - (wft.weight - start_of_frame_weight)
                if wft.flow > 0:
                    dt = dw / wft.flow
                    target_time = wft.weight_time + dt \
                                  - flow_sequencer.de1.stop_lead_time
                    if time.time() >= target_time:
                        await flow_sequencer.de1.skip_to_next()
                        self._last_frame_advanced_from = frame
                        mow_logger.info(
                            "Triggered frame "
                            "{:d} at {:.1f} g for {:.1f} g target".format(
                                frame, wft.weight, target))
                        await flow_sequencer.stop_at_notify(
                            stop_at=StopAtType.MOW,
                            action=StopAtNotificationAction.TRIGGERED,
                            target_value=target,
                            current_value=wafu.current_weight,
                            active_state=self.active_state,
                            current_frame=self.current_frame)
                        done = True

        return act_on_weight_subscriber

    async def stop_at_notify(self, stop_at: StopAtType,
                             action: StopAtNotificationAction,
                             target_value: Optional[float],
                             current_value: Optional[float],
                             active_state: API_MachineStates,
                             current_frame: Optional[int]):

        notification = StopAtNotification(
            stop_at=stop_at,
            action=action,
            target_value=target_value,
            current_value=current_value,
            active_state=active_state,
            current_frame=current_frame
        )
        # Using FlowSequencer() here still results in FlowSequencerImpl
        notification._sender = self
        await send_to_outbound_pipes(notification)

    async def auto_tare_notify(self, action: AutoTareNotificationAction):
        notification = AutoTareNotification(action=action)
        notification._sender = self
        await send_to_outbound_pipes(notification)

    def stop_at_time_set(self, state: API_MachineStates, duration: float):
        mode_control = self.active_control_for_state(state)
        if mode_control is None:
            raise DE1APIValueError(
                f"No ModeControl for {state}")
        if not isinstance(mode_control, StopAtTimeControl):
            raise DE1APIAttributeError(
                f"Does not implement StopAtTime {mode_control}"
            )
        try:
            if not mode_control.profile_can_override_stop_limits:
                return
        except AttributeError:
            pass
        mode_control.stop_at_time = duration

    def stop_at_volume_set(self, state: API_MachineStates, volume: float):
        mode_control = self.active_control_for_state(state)
        if mode_control is None:
            raise DE1APIValueError(
                f"No ModeControl for {state}")
        if not isinstance(mode_control, StopAtVolumeControl):
            raise DE1APIAttributeError(
                f"Does not implement StopAtVolume {mode_control}"
            )
        try:
            if not mode_control.profile_can_override_stop_limits:
                return
        except AttributeError:
            pass
        mode_control.stop_at_volume = volume

    def stop_at_weight_set(self, state: API_MachineStates, weight: float):
        mode_control = self.active_control_for_state(state)
        if mode_control is None:
            raise DE1APIValueError(
                f"No ModeControl for {state}")
        if not isinstance(mode_control, StopAtWeightControl):
            raise DE1APIAttributeError(
                f"Does not implement StopAtWeight {mode_control}"
            )
        try:
            if not mode_control.profile_can_override_stop_limits:
                return
        except AttributeError:
            pass
        mode_control.stop_at_weight = weight

    def profile_can_override_stop_limits(self, state: API_MachineStates):
        mode_control = self.active_control_for_state(state)
        if mode_control is None:
            raise DE1APIValueError(
                f"No ModeControl for {state}")
        try:
            return mode_control.profile_can_override_stop_limits
        except AttributeError:
            return True

    def profile_can_override_tank_temperature(self, state: API_MachineStates):
        mode_control = self.active_control_for_state(state)
        if mode_control is None:
            raise DE1APIValueError(
                f"No ModeControl for {state}")
        try:
            return mode_control.profile_can_override_tank_temperature
        except AttributeError:
            return True

    async def on_de1_nearly_ready(self):
        await self.hot_water_rinse_control.on_de1_nearly_ready()




class EspressoControl (I_EspressoControl):
    pass


class SteamControl (I_SteamControl):

    @property
    def stop_at_time(self):
        """
        Return the value from the DE1 for consistency with flush,
        which is managed by the DE1 starting with FW 1283
        """
        return FlowSequencer().de1._cuuid_dict[
            CUUID.ShotSettings].last_value.TargetSteamLength

    @stop_at_time.setter
    def stop_at_time(self, value):
        raise DE1APINotManagedHereException(
            "Steam time is set in the DE1 with CUUID.ShotSettings"
        )


class HotWaterControl (I_HotWaterControl):
    pass


class HotWaterRinseControl (I_HotWaterRinseControl):

    # Duration is controlled by FlowSequencer prior to FW 1283
    # but is handled by the DE1 after FW 1283
    # Define setter/getter that checks feature_flag.rinse_control

    # If the firmware version is unknown or is prior to 1283:
    #   manage the stop-at-time here
    #
    # If the DE1 is connected and is firmware enabled and a change is made:
    #   note that the change was made here and its value
    #   pass through to the DE1
    #   let the DE1 manage stop-at-time
    #
    # If a DE1 connects and is not firmware enabled:
    #   stay the course
    #
    # If a DE1 connects and is firmware enabled:
    #   And FS has been explicitly set through the API
    #       Use the FS value and send to the DE1
    #   But the FS has not been explicitly set
    #       Use the DE1 value, but don't mark as "explicitly set" (??)

    def __init__(self, *args, **kwargs):
        super(HotWaterRinseControl, self).__init__(*args, **kwargs)
        # Distinguish between not set and intentionally set to None
        self._stop_at_time_api_set = False

    @property
    def stop_at_time(self):
        de1 = FlowSequencer().de1
        if de1 is not None and de1.is_ready \
                and de1.feature_flag.rinse_control:
            retval = de1._mmr_dict[
                MMR0x80LowAddr.FLUSH_TIMEOUT].data_decoded
        else:
            retval = self._stop_at_time
        return retval

    @stop_at_time.setter
    def stop_at_time(self, value):
        raise DE1APINotManagedHereException(
            "Flush time needs to be set through the FlowSequencer "
            "as it may be managed by the DE1 with later firmware."
        )

    async def stop_at_time_set_async(self, value):
        self._stop_at_time = validate_stop_at(value)
        self._stop_at_time_api_set = True
        de1 = FlowSequencer().de1
        if de1 is not None and de1.is_ready \
                and de1.feature_flag.rinse_control:
            send_val = value
            if send_val is None:
                send_val = 0
            await de1.write_and_read_back_mmr0x80(
                MMR0x80LowAddr.FLUSH_TIMEOUT, send_val)
            new_val = de1._mmr_dict[
                MMR0x80LowAddr.FLUSH_TIMEOUT].data_decoded
            if new_val != send_val:
                logger.error(
                    f"Wrote flush, stop-at-time {send_val} "
                    f"but got back {new_val}")

    async def on_de1_nearly_ready(self):
        """
        Update based on what the DE1's MMR says and if locally set
        prior to connection.

        This conceivably could be handled with a callback on the MMR read
        but visions of infinite loops are dancing in my head.
        Going simple, if not completely robust, to start
        """
        de1 = FlowSequencer().de1
        if de1 is not None and de1.feature_flag.rinse_control:
            de1_val = de1._mmr_dict[
                MMR0x80LowAddr.FLUSH_TIMEOUT].data_decoded
            # 0 can mean "unimplemented MMR" or 0 or "ignore"
            # as feature_flag.rinse_control is true, take it as "ignore"
            if de1_val == 0:
                de1_val = None
            # If locally set
            if self._stop_at_time_api_set and de1_val != self._stop_at_time:
                # Override the DE1 setting
                await self.stop_at_time_set_async(self._stop_at_time)
            else:
                # Accept what the DE1 says
                # Don't use the setter as it already is in the DE1
                self._stop_at_time = de1_val
                self._stop_at_time_api_set = False


class WeightFlowTuple (NamedTuple):
    weight: float
    weight_time: float
    flow: float
    flow_time: float


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import email.utils
import json
import sqlite3
import time
# For the output of local_time
from datetime import datetime
from typing import NamedTuple, List, Union, Tuple

import aiosqlite

import pyDE1
import pyDE1.pyde1_logging as pyde1_logging
from pyDE1.exceptions import DE1IncompleteSequenceRecordError

logger = pyDE1.getLogger('LegacyShotFile')


class ShotRow (NamedTuple):
   de1_time:            float
   sample_time:         int
   group_pressure:      float
   group_flow:          float
   mix_temp:            float
   head_temp:           float
   set_mix_temp:        float
   set_head_temp:       float
   set_group_pressure:  float
   set_group_flow:      float
   frame_number:        int
   steam_temp:          float
   volume_preinfuse:    float
   volume_pour:         float
   volume_total:        float
   volume_by_frames:    str     # representation of list


def shot_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return ShotRow(*row)


class SequenceRow (NamedTuple):
    id: str
    active_state:   str
    start_sequence: float
    start_flow:     float
    end_flow:       float
    end_sequence:   float
    profile_id:     str
    # https://www.sqlite.org/quirks.html#no_separate_boolean_datatype
    profile_assumed:    int     # 0: False, 1: True
    resource_version:                           str
    resource_de1_id:                            str
    resource_de1_read_once:                     str
    resource_de1_calibration_flow_multiplier:   str
    resource_de1_control_mode:                  str
    resource_de1_control_tank_water_threshold:  str
    resource_de1_setting_before_flow:           str
    resource_de1_setting_steam:                 str
    resource_de1_setting_target_group_temp:     str
    resource_scale_id:                          str


def sequence_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return SequenceRow(*row)


class WeightRow (NamedTuple):
     current_weight:        float
     current_weight_time:   float


def weight_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return WeightRow(*row)


class FlowRow(NamedTuple):
    average_flow:       float
    average_flow_time:  float


def flow_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return FlowRow(*row)


class ProfileRow (NamedTuple):
    id:             str
    # source          bytes,
    source_format:  str
    fingerprint:    str
    date_added:     float
    title:          str
    author:         str
    notes:          str
    beverage_type:  str


def profile_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return ProfileRow(*row)


def braced_list(list_of_items: Union[List, Tuple]):
    return '{' + ' '.join(list_of_items) + '}'


async def legacy_shot_file(sequence_id: str,
                           db: aiosqlite.Connection):

    # contents will be joined with newlines prior to return
    contents = []
    # sequence_id is, or had better be, only hex digits with no spaces
    contents.append(f"sequence_id {sequence_id}")

    db.row_factory = sequence_row_factory
    cur = await db.execute(f"SELECT {' ,'.join(SequenceRow._fields)} "
                           f"FROM sequence WHERE id = ?", (sequence_id,))
    cur: aiosqlite.Cursor
    sequence_row: SequenceRow = await cur.fetchone()
    logger.debug(sequence_row)

    if sequence_row is None or None in (sequence_row.start_sequence,
                sequence_row.start_flow,
                sequence_row.end_flow,
                sequence_row.end_sequence):
        raise DE1IncompleteSequenceRecordError(
            f"Whoa there, not ready? {sequence_row})")

    contents.append(f"clock {round(sequence_row.start_flow)}")
    formatted_start_time = email.utils.format_datetime(
        datetime.fromtimestamp(sequence_row.start_flow).astimezone())
    contents.append(
        f"local_time {braced_list((formatted_start_time,))}")

    # Collect ShotSample data

    db.row_factory = shot_row_factory
    cur = await db.execute(
        f"SELECT {', '.join(ShotRow._fields)} "
        "FROM shot_sample_with_volume_update "
        "WHERE sequence_id = :sequence_id "
        "AND de1_time BETWEEN :start_flow AND :end_flow "
        "ORDER BY de1_time",
        (sequence_id, sequence_row.start_flow, sequence_row.end_flow))

    shot_rows: List[ShotRow] = await cur.fetchall()

    espresso_elapsed = [
        '{:.3f}'.format(r.de1_time - sequence_row.start_flow)
        for r in shot_rows
    ]
    contents.append(f"espresso_elapsed {braced_list(espresso_elapsed)}")

    espresso_pressure = [
        '{:.2f}'.format(r.group_pressure) for r in shot_rows
    ]
    contents.append(f"espresso_pressure {braced_list(espresso_pressure)}")

    espresso_flow = [
        '{:.2f}'.format(r.group_flow) for r in shot_rows
    ]
    contents.append(f"espresso_flow {braced_list(espresso_flow)}")

    espresso_temperature_basket = [
        '{:.2f}'.format(r.head_temp) for r in shot_rows
    ]
    contents.append("espresso_temperature_basket "
                    f"{braced_list(espresso_temperature_basket)}")

    espresso_temperature_mix = [
        '{:.2f}'.format(r.mix_temp) for r in shot_rows
    ]
    contents.append("espresso_temperature_mix "
                    f"{braced_list(espresso_temperature_mix)}")

    # This is scaled down by 10x in the legacy format
    espresso_water_dispensed = [
        '{:.2f}'.format(r.volume_total/10) for r in shot_rows
    ]
    contents.append(
        "WARNING {espresso_water_dispensed scaled by 10 "
        "for legacy compatibility}")
    contents.append("espresso_water_dispensed "
                    f"{braced_list(espresso_water_dispensed)}")

    espresso_temperature_goal = [
        '{:.2f}'.format(r.set_mix_temp) for r in shot_rows
    ]
    contents.append("espresso_temperature_goal "
                    f"{braced_list(espresso_temperature_goal)}")

    espresso_pressure_goal = [
        '{:.2f}'.format(r.set_group_pressure) for r in shot_rows
    ]
    contents.append("espresso_pressure_goal "
                    f"{braced_list(espresso_pressure_goal)}")

    espresso_flow_goal = [
        '{:.2f}'.format(r.set_group_flow) for r in shot_rows
    ]
    contents.append("espresso_flow_goal "
                    f"{braced_list(espresso_flow_goal)}")

    espresso_frame_number = [
        '{:d}'.format(r.frame_number) for r in shot_rows
    ]
    contents.append("espresso_frame_number "
                    f"{braced_list(espresso_frame_number)}")

    def bogus(frame_number: int):
        if frame_number % 2:
            return "10000000"
        else:
            return "-10000000"

    espresso_state_change = [
        bogus(r.frame_number) for r in shot_rows
    ]
    contents.append("espresso_state_change "
                    f"{braced_list(espresso_state_change)}")

    # About 2^-13 for a U16P12 value
    MIN_FLOW_FOR_CALC = 0.0001

    espresso_resistance = [
        '{:.2f}'.format(r.group_pressure /
                         max(r.group_flow, MIN_FLOW_FOR_CALC)**2)
        for r in shot_rows
    ]
    contents.append("espresso_resistance "
                    f"{braced_list(espresso_resistance)}")

    #
    # On to the scale
    #

    db.row_factory = weight_row_factory
    cur = await db.execute(
        f"SELECT {', '.join(WeightRow._fields)} "
        "FROM weight_and_flow_update "
        "WHERE sequence_id = :sequence_id "
        "AND current_weight_time BETWEEN :start_time AND :end_time "
        "ORDER BY current_weight_time",
        (sequence_id, sequence_row.start_flow, sequence_row.end_sequence))

    weight_rows: List[WeightRow] = await cur.fetchall()

    db.row_factory = flow_row_factory
    cur = await db.execute(
        f"SELECT {', '.join(FlowRow._fields)} "
        "FROM weight_and_flow_update "
        "WHERE sequence_id = :sequence_id "
        "AND average_flow_time BETWEEN :start_time AND :end_time "
        "ORDER BY average_flow_time",
        (sequence_id, sequence_row.start_flow, sequence_row.end_sequence))

    flow_rows: List[FlowRow] = await cur.fetchall()

    #
    # Legacy de1app behavior is to take the value just before the DE1 sample
    #

    # timeit on some variants here

    # O(n2) methods end up around 70 ms per series
    # list(filter(lambda r: r.current_weight_time < some_time, weight_rows)[-1]
    #   the last one is the "right" one, need to iterate all
    # next(filter(lambda r: r.current_weight_time >= some_time, weight_rows)
    #   the first one is the one just after, not before the time
    # O(n) method around 0.8 ms per series per execution (without formatting)
    # import timeit
    # n = 100
    # test_result = timeit.timeit(
    #     'emulate_ds(shot_rows, weight_rows, flow_rows)',
    #     globals=locals(),
    #     number=n)
    # print(f"{1000 * test_result / n:0.3f} ms per execution -- emulate_ds")

    def emulate_ds(shot: List[ShotRow],
                   weight: List[WeightRow],
                   flow: List[FlowRow]):
        idx_weight = 1
        idx_flow = 1
        last_w = len(weight) - 1
        last_f = len(flow) - 1
        retval_weight = []
        retval_flow = []

        for sr in shot:

            if last_w >= 1:
                while weight[idx_weight].current_weight_time < sr.de1_time \
                        and idx_weight < last_w:
                    idx_weight += 1
                retval_weight.append(
                    '{:.2f}'.format(weight[idx_weight-1].current_weight))

            if last_f >= 1:
                while flow[idx_flow].average_flow_time < sr.de1_time \
                        and idx_flow < last_f:
                    idx_flow += 1
                retval_flow.append(
                    '{:.2f}'.format(flow[idx_flow-1].average_flow))

        return retval_weight, retval_flow

    (espresso_weight, espresso_flow_weight) = emulate_ds(
        shot=shot_rows, weight=weight_rows, flow=flow_rows)


    contents.append("espresso_weight "
                    f"{braced_list(espresso_weight)}")

    contents.append("espresso_flow_weight "
                    f"{braced_list(espresso_flow_weight)}")

    # settings {
    #   drink_weight 12.3
    #   beverage_type
    # }
    # max of median flow after flow end, before end of sequence

    db.row_factory = None
    cur = await db.execute(
        f"SELECT max(median_weight) "
        "FROM weight_and_flow_update "
        "WHERE sequence_id = :sequence_id "
        "AND median_weight_time BETWEEN :start_time AND :end_time "
        "ORDER BY average_flow_time",
        (sequence_id, sequence_row.end_flow, sequence_row.end_sequence))

    (drink_weight,) = await cur.fetchone()

    db.row_factory = profile_row_factory
    cur = await db.execute(
        f"SELECT {', '.join(ProfileRow._fields)} "
        "FROM profile "
        "WHERE id = :id ",
        (sequence_row.profile_id,))

    profile_row: ProfileRow = await cur.fetchone()

    braced_name = braced_list((profile_row.title,))
    braced_bev_type = braced_list((profile_row.beverage_type,))

    contents.append("settings {")
    contents.append(f"\tprofile_title {braced_name}")
    contents.append(f"\tdrink_weight {drink_weight:0.1f}")
    contents.append(f"\tbeverage_type {braced_bev_type}")
    contents.append("}")

    # profile {
    #   (json)
    # }
    # Not used by Visualizer
    contents.append(f"profile {json.dumps(profile_row._asdict(), indent=2)}")

    return '\n'.join(contents) + "\n"


async def get_latest_espresso_id(db: aiosqlite.Connection) -> SequenceRow:

    db.row_factory = sequence_row_factory
    cur = await db.execute(f"SELECT {', '.join(SequenceRow._fields)} "
                           "FROM sequence "
                           "WHERE active_state == 'Espresso'"
                           "ORDER BY start_sequence DESC "
                           "LIMIT 1")
    return await cur.fetchone()


if __name__ == '__main__':

    import argparse

    from pyDE1.config import config

    ap = argparse.ArgumentParser(
        description="""Main executable to start the pyDE1 core.

        """
        f"Default configuration file is at {config.DEFAULT_CONFIG_FILE}"
    )
    ap.add_argument('-c', type=str, help='Use as alternate config file')

    args = ap.parse_args()

    pyde1_logging.setup_initial_logger()

    config.load_from_yaml(args.c)

    pyde1_logging.setup_direct_logging(config.logging)
    pyde1_logging.config_logger_levels(config.logging)

    async def run():
        async with aiosqlite.connect(config.database.FILENAME) as db:
            t0 = time.time()
            sr: SequenceRow = await get_latest_espresso_id(db)
            lsf = await legacy_shot_file(sr.id, db)
            t1 = time.time()
            print("-----")
            print(lsf, end='')
            print("-----")
            with open('test.shot', 'w') as fh:
                print(lsf, file=fh)
            print("Elapsed to retrieve and format: "
                  f"{(t1 - t0)*1000:0.1f} ms")

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    loop.run_until_complete(run())


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import json
import time
from typing import Optional

import aiosqlite

import pyDE1
from pyDE1.de1.profile import Profile
from pyDE1.event_manager import SequencerGateName
from pyDE1.event_manager.payloads import EventNotificationAction

logger = pyDE1.getLogger('Database.Insert')


async def profile(profile: Profile, db: aiosqlite.Connection, when: float):
    sql = "SELECT COUNT(*) FROM profile WHERE id == :id"
    cur = await db.execute(sql, {'id': profile.id})
    count = await cur.fetchone()
    if count[0] > 0:
        logger.info(f"Profile {profile.id} already in profile table.")
    else:
        sql = "INSERT INTO profile (id, source, source_format, fingerprint, " \
              "date_added, title, author, notes, beverage_type) " \
              "VALUES (:id, :source, :source_format, :fingerprint, " \
              ":date_added, :title, :author, :notes, :beverage_type)"
        vals = dict()
        for attr in ('id', 'source', 'source_format', 'fingerprint',
                     'title', 'author', 'notes', 'beverage_type'):
            vals[attr] = getattr(profile, attr)
        vals['date_added'] = when
        vals['source_format'] = vals['source_format'].value
        cur: aiosqlite.Cursor = await db.execute(sql, vals)
        await db.commit()
        logger.info(f"Profile {profile.id} added to profile table.")


async def persist_last_profile(profile: Profile, db: aiosqlite.Connection,
                               when: Optional[float] = None):
    if not when:
        when = time.time()
    # Update the last_profile information for persistence
    sql = "UPDATE persist_hkv " \
          "SET value=? " \
          "WHERE header='last_profile' AND key='id'"
    await db.execute(sql, (profile.id,))
    sql = "UPDATE persist_hkv " \
          "SET value=? " \
          "WHERE header='last_profile' AND key='datetime'"
    await db.execute(sql, (when,))
    await db.commit()


async def json_notification(notification: str,
                            sequence_id: str,
                            db: aiosqlite.Connection):
    as_dict = json.loads(notification)
    await  dict_notification(notification=as_dict,
                      sequence_id=sequence_id,
                      db=db)


async def dict_notification(notification: dict,
                            sequence_id: str,
                            db: aiosqlite.Connection):
    class_name = notification['class']
    try:
        method = CLASS_NAME_TO_METHOD[class_name]
        cur: aiosqlite.Cursor = await db.cursor()
        await method(notification=notification,
                     sequence_id=sequence_id,
                     cur=cur)
        await db.commit()
        await cur.close()
    except KeyError:
        logger.debug(f"No {__name__} method for {class_name}")


async def dict_notification_cursor_only(notification: dict,
                                        sequence_id: str,
                                        cur: aiosqlite.Cursor):
    class_name = notification['class']
    try:
        method = CLASS_NAME_TO_METHOD[class_name]
        await method(notification=notification,
                     sequence_id=sequence_id,
                     cur=cur)
    except KeyError:
        logger.debug(f"No {__name__} method for {class_name}")


async def shot_sample_with_volume_update(notification: dict,
                                         sequence_id: str,
                                         cur: aiosqlite.Cursor):
    sql = "INSERT INTO shot_sample_with_volume_update " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, de1_time, sample_time, group_pressure, " \
          "group_flow, mix_temp, head_temp, set_mix_temp, " \
          "set_head_temp, set_group_pressure, set_group_flow, " \
          "frame_number, steam_temp, volume_preinfuse, volume_pour, " \
          "volume_total, volume_by_frames) " \
          "VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :de1_time, :sample_time, :group_pressure, " \
          ":group_flow, :mix_temp, :head_temp, :set_mix_temp, " \
          ":set_head_temp, :set_group_pressure, :set_group_flow, " \
          ":frame_number, :steam_temp, :volume_preinfuse, :volume_pour, " \
          ":volume_total, :volume_by_frames)"
    notification['sequence_id'] = sequence_id
    notification['volume_by_frames'] = str(notification['volume_by_frames'])
    await cur.execute(sql, notification)


async def weight_and_flow_update(notification: dict,
                                 sequence_id: str,
                                 cur: aiosqlite.Cursor):
    sql = "INSERT INTO weight_and_flow_update " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, scale_time, current_weight, current_weight_time," \
          "average_flow, average_flow_time, " \
          "median_weight, median_weight_time, " \
          "median_flow, median_flow_time) " \
          "VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :scale_time, :current_weight, :current_weight_time," \
          ":average_flow, :average_flow_time, " \
          ":median_weight, :median_weight_time, " \
          ":median_flow, :median_flow_time)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


async def state_update(notification: dict,
                       sequence_id: str,
                       cur: aiosqlite.Cursor):
    sql = "INSERT INTO state_update " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, state, substate, " \
          "previous_state, previous_substate, is_error_state) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :state, :substate, " \
          ":previous_state, :previous_substate, :is_error_state)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


async def sequencer_gate_notification(notification: dict,
                                      sequence_id: str,
                                      cur: aiosqlite.Cursor):
    sql = "INSERT INTO sequencer_gate_notification " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, name, action, active_state) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :name, :action, :active_state)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)

    # Update the history record, if needed
    action = notification['action']
    name = notification['name']
    if action == EventNotificationAction.SET.value:
        target = None
        if name == SequencerGateName.GATE_FLOW_BEGIN.value:
            target = 'start_flow'
        elif name == SequencerGateName.GATE_FLOW_END.value:
            target = 'end_flow'
        elif name == SequencerGateName.GATE_SEQUENCE_COMPLETE.value:
            target = 'end_sequence'
        if target is not None:
            sql = f"UPDATE sequence SET {target} = ? " \
                  "WHERE sequence.id == ?"
            await cur.execute(sql, (notification['event_time'], sequence_id))


async def stop_at_notification(notification: dict,
                               sequence_id: str,
                               cur: aiosqlite.Cursor):
    sql = "INSERT INTO stop_at_notification " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, stop_at, action, active_state, target_value, " \
          "current_value) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :stop_at, :action, :active_state, :target_value, " \
          ":current_value)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


async def water_level_update(notification: dict,
                             sequence_id: str,
                             cur: aiosqlite.Cursor):
    sql = "INSERT INTO water_level_update " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, level, start_fill_level) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :level, :start_fill_level)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


async def scale_tare_seen(notification: dict,
                          sequence_id: str,
                          cur: aiosqlite.Cursor):
    sql = "INSERT INTO scale_tare_seen " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


async def auto_tare_notification(notification: dict,
                                 sequence_id: str,
                                 cur: aiosqlite.Cursor):
    sql = "INSERT INTO auto_tare_notification " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, action) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :action)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


async def scale_button_press(notification: dict,
                             sequence_id: str,
                             cur: aiosqlite.Cursor):
    sql = "INSERT INTO scale_button_press " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, button) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :button)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


async def connectivity_change(notification: dict,
                              sequence_id: str,
                              cur: aiosqlite.Cursor):
    sql = "INSERT INTO connectivity_change " \
          "(sequence_id, version, sender, arrival_time, create_time, " \
          "event_time, state, id, name) VALUES " \
          "(:sequence_id, :version, :sender, :arrival_time, :create_time, " \
          ":event_time, :state, :id, :name)"
    notification['sequence_id'] = sequence_id
    await cur.execute(sql, notification)


CLASS_NAME_TO_METHOD = {
    'ShotSampleWithVolumesUpdate': shot_sample_with_volume_update,
    'WeightAndFlowUpdate': weight_and_flow_update,
    'StateUpdate': state_update,
    'SequencerGateNotification': sequencer_gate_notification,
    'StopAtNotification': stop_at_notification,
    'WaterLevelUpdate': water_level_update,
    'ScaleTareSeen': scale_tare_seen,
    'AutoTareNotification': auto_tare_notification,
    'ScaleButtonPress': scale_button_press,
    'ConnectivityChange': connectivity_change,
}


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Provides pyDE1.getLogger()

"""
import logging
from typing import Optional

_ROOT_LOGGER_NAME = 'pyDE1'
_ROOT_LOGGER_PREFIX = _ROOT_LOGGER_NAME + '.'
_ROOT_LOGGER_PREFIX_LEN = len(_ROOT_LOGGER_PREFIX)


def getLogger(name: Optional[str]=None) -> logging.Logger:
    pyde1_root = logging.getLogger(_ROOT_LOGGER_NAME)
    retval = None
    if name is None:
        retval = pyde1_root
    elif name == 'root':
        retval = logging.getLogger()
    elif name.startswith('root.'):
        retval = logging.getLogger(name[5:])
    else:
        retval = pyde1_root.getChild(name)
    return retval